#!/usr/bin/env python

from __future__ import division

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
LINE_STYLE = ['b:+', 'g-', 'r-s', 'c--', 'k-.']

import traceback
import argparse
import random
import time
import numpy as np
from collections import defaultdict
from csv import DictWriter
from util.stats import proportion_ci

from multiprocessing import Pool, cpu_count
from schedcat.util.storage import storage
from generator.generator import DesignPointGenerator

from distributions import *

from schedcat.model.LLC_params import *
from schedcat.generator.mcx_tasks import TaskGeneratorMCX
from LLCOptimizer import *
from LLCCanvas import LLCCanvas
import schedcat.model.ohc as ohc
import loadohs

from sched_check import *

MIN_SAMPLES = 100
MAX_SAMPLES = 2000
MAX_CI      = 0.05
CONFIDENCE  = 0.95

topdir = "~/LLC_Code/MCXSchedStudy/"

def make_mcx_task_set(dp, to_plot = False):

    tgMCX = TaskGeneratorMCX(
                             period = NAMED_PERIODS[dp.period],
                             util = NAMED_UTILIZATIONS[dp.task_util],
                             util_dist = NAMED_UTIL_DISTS[dp.util_dist],
                             mem_util = NAMED_MEMORY[dp.memory_util],
                             avg_percent = AVG_PERCENT,
                             a_inflation = NAMED_A_INFLATIONS[dp.inflation_A],
                             pfuncs = P)
        
    tsMCX = tgMCX.make_task_set(max_util = dp.sys_util)
    
    tsMCX.m = CORES
    tsMCX.wg = 1
    tsMCX.scheme = bypass_scheme(CORES)
                             
    return tsMCX

conf_keys = ['MILP', 'LP']

def complete(results, n):
    if n < MIN_SAMPLES:
        return False
    elif n > MAX_SAMPLES:
        return True
    else:
        for key in conf_keys:
            if proportion_ci(results[key], n, CONFIDENCE) > MAX_CI:
                return False
        return True

def update_mean(old_mean, n, new_sample):
    return (old_mean*n + new_sample)/(n+1)


def process_dp(dp):
    sched_res = defaultdict(float)
    count = defaultdict(int)
    ps_hist = defaultdict(float) #np.zeros(51)
    
    n = 0
    sched_res['Time-Outs'] = 0
    sched_res['MILP-Time-Max'] = 0.0
    sched_res['LP-Time-Max'] = 0.0
    while not complete(sched_res, n):
        
        ts = make_mcx_task_set(dp)
        
        unmnwts = []
        loadwtsAB = []
        loadwtsC = []
        
        for tsk in ts:
            unmnwts.append(unmanwt())
            loadwtsAB.append(loadedinfAB())
            
        for tsk in ts.level('C'):
                loadwtsC.append(loadedinfC())
        
        ts.useCshared()
        
        #---------inflate for loads in MC-iso------------#
        for idx, tsk in enumerate(ts):
            if tsk.level is 'C':
                break
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] *= loadwtsAB[idx]    
                        
        for idx, tsk in enumerate(ts.level('C')):
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] *= loadwtsC[idx]  
        #------------------------------------------------#
        
        ts.setLLCAlloc(bypass_scheme(ts.m))
        ohc.SetOheads(ts, 'mc2') 
        
        ts.coreAssign(useMC2 = True)
        
        minAperiods = []
        for core in range(0, 4):
            if len(ts.level('A').core(core)) == 0:
                minAperiods.append(0)
            else:
                minAperiods.append(min([t.period for t in ts.level('A').core(core)]))
        
        start = time.clock() 
        psMILP, t = LPLLCways(ts, 'Integer', csch = default_scheme(ts.m), olap = True, debug = False)
        if psMILP == None:
            psMILP, t = LPLLCways(ts, 'Integer', csch = default_scheme(ts.m), olap = False, debug = False)
        fin = time.clock() 
        
        MILP_Time = float(fin-start)
        
        to = 0
        if t == -1:
            to = 1
        
        schedMILP = psMILP != None
        
        start = time.clock() 
        psLP, t = LPLLCways(ts, 'Continuous', csch = default_scheme(ts.m), olap = True, debug = False)
        if psLP == None:
            psLP, t = LPLLCways(ts, 'Continuous', csch = default_scheme(ts.m), olap = False, debug = False)
        fin = time.clock() 
        
        LP_Time = float(fin-start)
        
        schedLP = psLP != None
        
        sched_res['MILP'] = update_mean(sched_res['MILP'], n, schedMILP)
        sched_res['LP']   = update_mean(sched_res['LP'],   n, schedLP)
        
        sched_res['MILP-Time-Avg'] = update_mean(sched_res['MILP-Time-Avg'], n, MILP_Time)
        sched_res['LP-Time-Avg']   = update_mean(sched_res['LP-Time-Avg'],   n, LP_Time)
        
        sched_res['MILP-Time-Max'] = max   (sched_res['MILP-Time-Max'], MILP_Time)
        sched_res['LP-Time-Max']   = max   (sched_res['LP-Time-Max'],   LP_Time)
        
        sched_res['Time-Outs'] = sched_res['Time-Outs']+to
        
        count['Count'] = n+1

        n+=1
    
    return dict(dp.items() + sched_res.items()+count.items())

def myrange(start, end, inc):
    return np.arange(start, end+inc, inc)
    
def valid(dp):
    return True

def thread_tests(args, outs, design_points):
    pool = Pool(processes = args.processors)
    try:
        for i, row in enumerate(pool.imap_unordered(process_dp, design_points)):
            if not row:
                continue
            sys.stderr.write('\rdone {0:%}'.format(i/len(design_points)))
    
            title = titlefor(row)
            outs[title].writerow(row)
        pool.close()
    except Exception as e:
        _, _, tb = sys.exc_info()
        tb_info = traceback.extract_tb(tb)
        _, _, _, text = tb_info[-1]
        print "hit an exception"
        print e
        print "text:"
        print text
        pool.terminate()
        pool.join()

def titlefor(headers):
    return headers['period'] + "--" + headers['task_util'] + "--" + headers['inflation_A'] + "--" + headers['memory_util'] + "--" + headers['util_dist']

def main():
    parser = argparse.ArgumentParser()
    # parser.add_argument('-d', "--datafile", type = argparse.FileType('w'),
                        # default = sys.stdout,
                        # help = "File to output csv file to")
    parser.add_argument('-p', "--plot", action='store_true',
                        help = "For plotting generated PETs")
    parser.add_argument('-m', "--processors", default=cpu_count()-5, type = int,
                        help = "Number of processors to execute on")
    args = parser.parse_args()
                        
    exp = storage()

    exp.task_util = NAMED_UTILIZATIONS.keys()
    exp.inflation_A = NAMED_A_INFLATIONS.keys()
    exp.period = NAMED_PERIODS.keys()
    exp.memory_util = NAMED_MEMORY.keys()
    exp.util_dist = NAMED_UTIL_DISTS.keys()
    exp.sys_util = myrange(0.1, 9.1, 0.2)
    
    design_points = [dp for dp in DesignPointGenerator(exp, is_valid = valid)]

    random.seed(43215)
    random.shuffle(design_points)
    
    loadohs.load("../oheads/")
    
    print "Total design points: ", len(design_points)
    
    outs = dict()
    for dp in design_points:
        title = titlefor(dp)
        if not title in outs:
            outs[title] = DictWriter(open("../results_LP_V_MILP/" + title, 'w'), exp.keys()+['MILP', 'LP', 'MILP-Time-Avg', 'LP-Time-Avg', 'MILP-Time-Max', 'LP-Time-Max', 'Time-Outs', 'Count'])
            outs[title].writeheader()

#     for i, t in enumerate(design_points):
#         process_dp(design_points[i])

    thread_tests(args, outs, design_points)

if __name__ == '__main__':
    main()