#!/usr/bin/env python

from __future__ import division

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
LINE_STYLE = ['b:+', 'g-', 'r-s', 'c--', 'k-.']

import argparse
import random
import time
import numpy as np
from collections import defaultdict
from csv import DictWriter
from util.stats import proportion_ci

from multiprocessing import Pool, cpu_count
from schedcat.util.storage import storage
from generator.generator import DesignPointGenerator

from distributions import *

from schedcat.model.LLC_params import *
from schedcat.generator.mcx_tasks import TaskGeneratorMCX
from LLCOptimizer import *
from LLCCanvas import LLCCanvas
import schedcat.model.ohc as ohc
import loadohs

from sched_check import *

MIN_SAMPLES = 100
MAX_SAMPLES = 2000
MAX_CI      = 0.05
CONFIDENCE  = 0.95

topdir = "~/LLC_Code/MCXSchedStudy/"

def make_mcx_task_set(dp, to_plot = False):

    tgMCX = TaskGeneratorMCX(
                             period = NAMED_PERIODS[dp.period],
                             util = NAMED_UTILIZATIONS[dp.task_util],
                             util_dist = NAMED_UTIL_DISTS[dp.util_dist],
                             mem_util = NAMED_MEMORY[dp.memory_util],
                             avg_percent = AVG_PERCENT,
                             a_inflation = NAMED_A_INFLATIONS[dp.inflation_A],
                             pfuncs = P)
        
    tsMCX = tgMCX.make_task_set(max_util = dp.sys_util)
    
    tsMCX.m = CORES
    tsMCX.wg = 1
    tsMCX.scheme = bypass_scheme(CORES)
                             
    return tsMCX

def complete(results, n):
    if n < MIN_SAMPLES:
        return False
    elif n > MAX_SAMPLES:
        return True
    else:
        for key in results:
            if proportion_ci(results[key], n, CONFIDENCE) > MAX_CI:
                return False
        return True

def update_mean(old_mean, n, new_sample):
    return (old_mean*n + new_sample)/(n+1)


def process_dp(dp):
    sched_res = defaultdict(float)
    count = defaultdict(int)
    ps_hist = defaultdict(float) #np.zeros(51)
    
    n = 0
    
    while not complete(sched_res, n):
        
        ts = make_mcx_task_set(dp)
        
        unmnwts = []
        loadwtsAB = []
        loadwtsC = []
        
        for tsk in ts:
            unmnwts.append(unmanwt())
            loadwtsAB.append(loadedinfAB())
            
        for tsk in ts.level('C'):
                loadwtsC.append(loadedinfC())
        
        ts.useCshared()
        
        #---------inflate for loads in MC-iso------------#
        for idx, tsk in enumerate(ts):
            if tsk.level is 'C':
                break
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] *= loadwtsAB[idx]    
                        
        for idx, tsk in enumerate(ts.level('C')):
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] *= loadwtsC[idx]  
        #------------------------------------------------#
        
        ts.setLLCAlloc(bypass_scheme(ts.m))
        ohc.SetOheads(ts, 'mc2') 
        
        ts.coreAssign(useMC2 = True)
        
        minAperiods = []
        for core in range(0, 4):
            if len(ts.level('A').core(core)) == 0:
                minAperiods.append(0)
            else:
                minAperiods.append(min([t.period for t in ts.level('A').core(core)]))
            
        psMC2Iso, MC2IsoTime = LPLLCways(ts, 'Integer', csch = default_scheme(ts.m), olap = True, debug = False)
        if psMC2Iso == None:
            psMC2Iso, MC2IsoTime = LPLLCways(ts, 'Integer', csch = default_scheme(ts.m), olap = False, debug = False)
        
        schedMC2Iso = psMC2Iso != None
        
        psMC2IsoLP, MC2IsoTimeLP = LPLLCways(ts, 'Continuous', csch = default_scheme(ts.m), olap = True, debug = False)
        if psMC2IsoLP == None:
            psMC2IsoLP, MC2IsoTimeLP = LPLLCways(ts, 'Continuous', csch = default_scheme(ts.m), olap = False, debug = False)
        
        schedMC2IsoLP = psMC2IsoLP != None
        
        
        for idx, tsk in enumerate(ts):
            for s in range(0, 17):
                for w in range(0, 17):
                    if tsk.level is 'A' or tsk.level is 'B':
                        tsk.cost[crt][s][w] *= unmnwts[idx]
        
        ts.coreAssign(useMC2 = True)
        psMC2ManLLC, MC2ManLLCTime = LPLLCways(ts, 'Integer', csch = default_scheme(ts.m), olap = True, debug = False)
        if psMC2ManLLC == None:
            psMC2ManLLC, MC2ManLLCTime = LPLLCways(ts, 'Integer', csch = default_scheme(ts.m), olap = False, debug = False)
        
        schedMC2ManLLC = psMC2ManLLC != None #LLC isolation, no bank isolation
        
        for idx, tsk in enumerate(ts):
            for s in range(0, 17):
                for w in range(0, 17):
                    if tsk.level is 'A' or tsk.level is 'B':
                        tsk.cost[crt][s][w] /= unmnwts[idx]
        
        
        #---------undo inflates for loads in MC-iso------------#
        for idx, tsk in enumerate(ts):
            if tsk.level is 'C':
                break
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] /= loadwtsAB[idx]    
                        
        for idx, tsk in enumerate(ts.level('C')):
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] /= loadwtsC[idx]  
        #------------------------------------------------#
        
        ts.useCiso()
        
        #---------inflate for loads in MC-no-C-share------------#
        for idx, tsk in enumerate(ts):
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] *= loadwtsAB[idx]    
        #-------------------------------------------------------#
        
        ts.coreAssign(useMC2 = True)
        
        psMC2_No_Share, MC2_No_ShareTime = LPLLCways(ts, 'Integer', csch = c_iso_scheme(ts.m), olap = True, debug = False)
        if psMC2_No_Share == None:
            psMC2_No_Share, MC2_No_ShareTime = LPLLCways(ts, 'Integer', csch = c_iso_scheme(ts.m), olap = False, debug = False)
        
        schedMC2_No_Share = psMC2_No_Share != None
        #---------undo inflate for loads in MC-no-C-share-------#
        for idx, tsk in enumerate(ts):
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] /= loadwtsAB[idx]    
        #--------------------------------------------------------#
        
        # if schedMC2Iso and not is_sched(ts, psMC2Iso, debug = False):
#             print "inconsistency1 " + str(dp.sys_util) + " " + str(dp.task_util) + " " + str(dp.period) + " " + str(dp.memory_util) + " " + dp.util_dist
#             for i in range(0, 4):
#                 print "partition on core " + str(i)
#                 print "  " + str(psMC2Iso.A[i].W) + ", " + str(psMC2Iso.A[i].S)
#                 print "  " + str(psMC2Iso.B[i].W) + ", " + str(psMC2Iso.B[i].S)
#             
#             print str(psMC2Iso.C.W) + ", " + str(psMC2Iso.C.S)
#             is_sched(ts, psMC2Iso, debug = True)
#             LPLLCways(ts, 'Integer', csch = default_scheme(), olap = True, debug = True)
#             if psMC2Iso == None:
#                 print "TRYING NO OVERLAP"
#                 LPLLCways(ts, 'Integer', csch = default_scheme(), olap = False, debug = True)
# #             exit(0)
        
        ts.useCshared()
        
        #---------inflate for loads in MC-Cshared------------#
        for idx, tsk in enumerate(ts):
            if tsk.level is 'C':
                break
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] *= loadwtsAB[idx]    
                        
        for idx, tsk in enumerate(ts.level('C')):
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] *= loadwtsC[idx]  
        #------------------------------------------------#
        
        ts.coreAssign(useMC2 = True)
        schedMC2UnmanBank = is_sched(ts, bypass_scheme(ts.m), debug = False)
        
        for idx, tsk in enumerate(ts):
            for s in range(0, 17):
                for w in range(0, 17):
                    if tsk.level is 'A' or tsk.level is 'B':
#                     for crt in {'A', 'B'}:
                        tsk.cost[crt][s][w] *= unmnwts[idx]
        
        ts.coreAssign(useMC2 = True)
        schedMC2Unman = is_sched(ts, bypass_scheme(ts.m), debug = False)
            
        for idx, tsk in enumerate(ts):
            for s in range(0, 17):
                for w in range(0, 17):
                    if tsk.level is 'A' or tsk.level is 'B':
#                     for crt in {'A', 'B'}:
                        tsk.cost[crt][s][w] /= unmnwts[idx]
        
        #---------undo inflates for loads in MC-Cshared------------#
        for idx, tsk in enumerate(ts):
            if tsk.level is 'C':
                break
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] /= loadwtsAB[idx]    
                        
        for idx, tsk in enumerate(ts.level('C')):
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] /= loadwtsC[idx]  
        #------------------------------------------------#
        
        
        ts.useCiso()
        
        
        for tsk in ts:
            tsk.wasc = tsk.level is 'C'
            tsk.level = 'B'
        
        for tsk in ts:
            for s in range(0, 17):
                for w in range(0, 17):
                    tsk.cost['B'][s][w] = tsk.cost['A'][s][w]    
        
        
        #---------inflate for loads in MC-no-C-share------------#
        for idx, tsk in enumerate(ts):
            for s in range(0, 17):
                for w in range(0, 17):
                    tsk.cost['B'][s][w] *= loadwtsAB[idx]    
        #-------------------------------------------------------#
        
        ts.setLLCAlloc(pedf_scheme(ts.m))
        ohc.SetOheads(ts, 'pedf')
        ts.coreAssign(useMC2 = False)

        psPEDFIso, PEDFIsoTime = LPLLCsets(ts, 'Integer', csch = pedf_scheme(ts.m), olap = False)
        
        schedPEDFIso = psPEDFIso != None  
        
#         if schedPEDFIso and not is_sched(ts, psPEDFIso, debug = False):
#             print "inconsistency2 " + str(dp.sys_util) + " " + str(dp.task_util) + " " + str(dp.period) + " " + str(dp.memory_util) + " " + dp.util_dist
#             is_sched(ts, psPEDFIso, debug = True)
#             exit(0)
        
        for idx, tsk in enumerate(ts):
            if not tsk.wasc:
                for s in range(0, 17):
                    for w in range(0, 17):
                        tsk.cost['B'][s][w] *= unmnwts[idx]
        
        ts.coreAssign(useMC2 = True)
        schedPEDFUnman = is_sched(ts, bypass_scheme(ts.m), debug = False)  
        
        for idx, tsk in enumerate(ts):
            if not tsk.wasc:
                for s in range(0, 17):
                    for w in range(0, 17):
                        tsk.cost['B'][s][w] /= unmnwts[idx]
                   
        #---------undo-inflate for loads in MC-no-C-share------------#
        for idx, tsk in enumerate(ts):
            for s in range(0, 17):
                for w in range(0, 17):
                    tsk.cost['B'][s][w] /= loadwtsAB[idx]
        #------------------------------------------------------------#
        
        ts.setScheme(uni_scheme(ts.m))
        mx_wss_all = max([tsk.wss for tsk in ts])
        
        mx_wss = {'B' : [mx_wss_all, mx_wss_all, mx_wss_all, mx_wss_all],
                  'C' : 0}
        
        ohc.CPMD = 30
        schedEDFUni = ts.utilization('B', approximate = False, inflated = True, c_taskset = [], minAperiods = [0, 0, 0, 0], mx_wss = mx_wss) <= 1
                
        sched_res['MC2-Iso']       = update_mean(sched_res['MC2-Iso'],       n, schedMC2Iso)
        sched_res['MC2-Iso-LP']    = update_mean(sched_res['MC2-Iso-LP'],    n, schedMC2IsoLP)
        sched_res['MC2-ManLLC']    = update_mean(sched_res['MC2-ManLLC'],    n, schedMC2ManLLC)
        sched_res['PEDF-Iso']      = update_mean(sched_res['PEDF-Iso'],      n, schedPEDFIso)
        sched_res['MC2-Unman']     = update_mean(sched_res['MC2-Unman'],     n, schedMC2Unman)
        sched_res['MC2-UnmanBank'] = update_mean(sched_res['MC2-UnmanBank'], n, schedMC2UnmanBank) #bank isolation, no LLC isolation
        sched_res['PEDF-Unman']    = update_mean(sched_res['PEDF-Unman'],    n, schedPEDFUnman)
        sched_res['EDF-Uni']       = update_mean(sched_res['EDF-Uni'],       n, schedEDFUni)
        sched_res['MC2-No-Share']  = update_mean(sched_res['MC2-No-Share'],  n, schedMC2_No_Share)
        
        
        sched_res['MC2Iso-Time-Avg']     = update_mean(sched_res['MC2Iso-Time-Avg'],     n, MC2IsoTime)
        sched_res['MC2IsoLP-Time-Avg']   = update_mean(sched_res['MC2IsoLP-Time-Avg'],   n, MC2IsoTimeLP)
        sched_res['MC2ManLLC-Time-Avg']  = update_mean(sched_res['MC2ManLLC-Time-Avg'],  n, MC2ManLLCTime)
        sched_res['PEDFIso-Time-Avg']    = update_mean(sched_res['PEDFIso-Time-Avg'],    n, PEDFIsoTime)
        sched_res['MC2NoShare-Time-Avg'] = update_mean(sched_res['MC2NoShare-Time-Avg'], n, MC2_No_ShareTime)
        
        
#         try:
#         fmt = "{:2.4f}"
#         print fmt.format(MC2IsoTime),    fmt.format(MC2IsoTimeLP), fmt.format(MC2ManLLCTime), fmt.format(PEDFIsoTime), fmt.format(MC2_No_ShareTime), "    ", fmt.format(sched_res['MC2Iso-Time-Max']),    fmt.format(sched_res['MC2IsoLP-Time-Max']), fmt.format(sched_res['MC2ManLLC-Time-Max']), fmt.format(sched_res['PEDFIso-Time-Max']), fmt.format(sched_res['MC2NoShare-Time-Max'])
              
        sched_res['MC2Iso-Time-Max']     = 0#my_max(sched_res['MC2Iso-Time-Max'],     MC2IsoTime)
        sched_res['MC2IsoLP-Time-Max']   = 0#my_max(sched_res['MC2IsoLP-Time-Max'],   MC2IsoTimeLP)
        sched_res['MC2ManLLC-Time-Max']  = 0#my_max(sched_res['MC2ManLLC-Time-Max'],  MC2ManLLCTime)
        sched_res['PEDFIso-Time-Max']    = 0#my_max(sched_res['PEDFIso-Time-Max'],    PEDFIsoTime)
        sched_res['MC2NoShare-Time-Max'] = 0#my_max(sched_res['MC2NoShare-Time-Max'], MC2_No_ShareTime)
#         except:
#             _, _, tb = sys.exc_info()
#             tb_info = traceback.extract_tb(tb)
#             _, _, _, text = tb_info[-1]
#             print "Error", text
#             raise
        
        count['Count'] = n+1

        n+=1
    
    return dict(dp.items() + sched_res.items()+count.items())

def my_max(a, b):
    if a > b:
        return float(b)
    else:
        return float(b)

def myrange(start, end, inc):
    return np.arange(start, end+inc, inc)
    
def valid(dp):
    return True

Result_Headers = ['MC2-Iso',
                  'MC2-Iso-LP', 
                  'MC2-ManLLC', 
                  'PEDF-Iso', 
                  'MC2-Unman', 
                  'PEDF-Unman', 
                  'EDF-Uni', 
                  'MC2-No-Share', 
                  'MC2-UnmanBank', 
                  'MC2Iso-Time-Avg',
                  'MC2IsoLP-Time-Avg',
                  'MC2ManLLC-Time-Avg',
                  'PEDFIso-Time-Avg',
                  'MC2NoShare-Time-Avg',
                  'MC2Iso-Time-Max',
                  'MC2IsoLP-Time-Max',
                  'MC2ManLLC-Time-Max',
                  'PEDFIso-Time-Max',
                  'MC2NoShare-Time-Max',
                  'Count']

def thread_tests(args, outs, design_points):
    pool = Pool(processes = args.processors)
    try:
        for i, row in enumerate(pool.imap_unordered(process_dp, design_points)):
            if not row:
                continue
            sys.stderr.write('\rdone {0:%}'.format(i/len(design_points)))
    
            title = titlefor(row)
            outs[title].writerow(row)
        pool.close()
    except Exception as e:
        print "hit an exception"
        print e
        pool.terminate()
        pool.join()

def titlefor(headers):
    return headers['period'] + "--" + headers['task_util'] + "--" + headers['inflation_A'] + "--" + headers['memory_util'] + "--" + headers['util_dist']

def main():
    parser = argparse.ArgumentParser()
    # parser.add_argument('-d', "--datafile", type = argparse.FileType('w'),
                        # default = sys.stdout,
                        # help = "File to output csv file to")
    parser.add_argument('-p', "--plot", action='store_true',
                        help = "For plotting generated PETs")
    parser.add_argument('-m', "--processors", default=cpu_count(), type = int,
                        help = "Number of processors to execute on")
    args = parser.parse_args()
                        
    exp = storage()

    exp.task_util = NAMED_UTILIZATIONS.keys()
    exp.inflation_A = NAMED_A_INFLATIONS.keys()
    exp.period = NAMED_PERIODS.keys()
    exp.memory_util = NAMED_MEMORY.keys()
    exp.util_dist = NAMED_UTIL_DISTS.keys()
    exp.sys_util = myrange(0.1, 9.1, 0.2)
    
    design_points = [dp for dp in DesignPointGenerator(exp, is_valid = valid)]

    random.seed(43215)
    # random.shuffle(design_points)
    
    loadohs.load("../oheads/")

    if args.plot:
        print "plotting"
        
        ts = make_mcx_task_set(design_points[0])
        
        ts.useCiso()
        ts.setLLCAlloc(bypass_scheme())
        ohc.SetOheads(ts, 'mc2') 
        ts.coreAssign(useMC2 = True)
        
#         LLCcan = LLCCanvas(default_scheme(), ts)
#         ts[0].plot('C')
        
        # ts.plot('C', 'C', 'ex-exec-ts-iso.pdf')
#         ts.useCshared()
#         ts.plot('C', 'C', 'ex-exec-ts-shared.pdf')

        ts.level('C')[1].plot2()
#         ts.plot('C', 'C', 'ex-exec-ts.pdf')
        
        sys.exit(0)
    
    print "Total design points: ", len(design_points)
    # if(len(os.listdir(os.getcwd() + "/results/")) > 0):
        # print "already have results"
        # exit()
    
    outs = dict()
    for dp in design_points:
        title = titlefor(dp)
        if not title in outs:
            outs[title] = DictWriter(open("../results_final_rev/" + title, 'w'), exp.keys()+Result_Headers)
            outs[title].writeheader()

#     for i, t in enumerate(design_points):
#         process_dp(design_points[i])

    thread_tests(args, outs, design_points)

if __name__ == '__main__':
    main()