#!/usr/bin/env python

from __future__ import division

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
LINE_STYLE = ['b:+', 'g-', 'r-s', 'c--', 'k-.']

import argparse
import random
import time
import numpy as np
from collections import defaultdict
from csv import DictWriter
from util.stats import proportion_ci

from multiprocessing import Pool, cpu_count, Manager
from schedcat.util.storage import storage
from generator.generator import DesignPointGenerator

from distributions import *

from schedcat.model.LLC_params import *
from schedcat.generator.mcx_tasks import TaskGeneratorMCX
from LLCOptimizer import *
from LLCCanvas import LLCCanvas
import schedcat.model.ohc as ohc
import loadohs

from sched_check import *

# NUM_SAMPLES = 5
MIN_SAMPLES = 100
MAX_SAMPLES = 2000
MAX_CI      = 0.05
CONFIDENCE  = 0.95

topdir = "~/LLC_Code/MCXSchedStudy/"

def make_mcx_task_set(dp, to_plot = False):

    tgMCX = TaskGeneratorMCX(
                             period = NAMED_PERIODS[dp.period],
                             util = NAMED_UTILIZATIONS[dp.task_util],
                             util_dist = NAMED_UTIL_DISTS[dp.util_dist],
                             mem_util = NAMED_MEMORY[dp.memory_util],
                             avg_percent = AVG_PERCENT,
                             a_inflation = NAMED_A_INFLATIONS[dp.inflation_A],
                             pfuncs = P,
                             way_gran = dp.way_gran)
        
    tsMCX = tgMCX.make_task_set(max_util = dp.sys_util)
    
    tsMCX.m = int(dp.cores)
    tsMCX.wg = int(dp.way_gran)
    tsMCX.scheme = bypass_scheme(tsMCX.m)
                             
    return tsMCX
    
def complete(results, n):
    if n < MIN_SAMPLES:
        return False
    elif n >= MAX_SAMPLES:
        return True
    elif proportion_ci(results['sched'], n, CONFIDENCE) > MAX_CI:
        return False
    else:
        return True

def update_mean(old_mean, n, new_sample):
    return (old_mean*n + new_sample)/(n+1)

def process_dp(dp):
    sched_res = defaultdict(float)
    count = defaultdict(int)
    ps_hist = defaultdict(float) #np.zeros(51)
    
    sched_res['time'] = []
    sched_res['solve_time'] = []
    
    n = 0
    while not complete(sched_res, n):
        ts = make_mcx_task_set(dp)
        
        unmnwts = []
        loadwtsAB = []
        loadwtsC = []
        
        for tsk in ts:
            unmnwts.append(unmanwt())
            loadwtsAB.append(loadedinfAB())
            
        for tsk in ts.level('C'):
                loadwtsC.append(loadedinfC())
        
        ts.useCshared()
        
        #---------inflate for loads in MC-iso------------#
        for idx, tsk in enumerate(ts):
            if tsk.level is 'C':
                break
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] *= loadwtsAB[idx]    
                        
        for idx, tsk in enumerate(ts.level('C')):
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] *= loadwtsC[idx]  
        #------------------------------------------------#
        
        ts.setLLCAlloc(bypass_scheme(ts.m))
        ohc.SetOheads(ts, 'mc2') 
        
        ts.coreAssign(useMC2 = True)
        
        minAperiods = []
        for core in range(0, ts.m):
            if len(ts.level('A').core(core)) == 0:
                minAperiods.append(0)
            else:
                minAperiods.append(min([t.period for t in ts.level('A').core(core)]))
           
        start = time.clock() 
        psMC2Iso, slv_t = LPLLCways(ts, 'Integer', csch = default_scheme(ts.m), olap = True, debug = False)
        if psMC2Iso == None:
            psMC2Iso, slv_t = LPLLCways(ts, 'Integer', csch = default_scheme(ts.m), olap = False, debug = False)
        
        schedMC2Iso = psMC2Iso != None and psMC2Iso != "TimeOut"
        fin = time.clock()
        
        sched_res['sched'] = update_mean(sched_res['sched'],    n, schedMC2Iso)
        
        
        if psMC2Iso == "TimeOut":
            sched_res['time'].append(-1)
        else:
            sched_res['time'].append(fin-start)
        
        sched_res['solve_time'].append(slv_t)
        n+=1
        
    for rec in range (n, MAX_SAMPLES):
        sched_res['time'].append(-2)
        sched_res['solve_time'].append(-2)
    
    return dict(dp.items() + sched_res.items())

def myrange(start, end, inc):
    return np.arange(start, end+inc, inc)
    
def valid(dp):
    return True

def proc_result(outs_sch, times, solve_times, result):
    if not result:
        return
    
    title = titlefor(result)

    outs_sch[title].writerow({'sys_util': result['sys_util'], 'sched': result['sched']})

    for t_ndx, e in enumerate(result['time']):
        samples = times[title]
        utils = samples[t_ndx]
        utils[result['sys_util']] = e
        samples[t_ndx] = utils
        times[title] = samples
        
    for t_ndx, e in enumerate(result['solve_time']):
        samples = solve_times[title]
        utils = samples[t_ndx]
        utils[result['sys_util']] = e
        samples[t_ndx] = utils
        solve_times[title] = samples
    
def thread_tests(args, outs_sch, times, solve_times, design_points):
    pool = Pool(processes = args.processors)
    try:
        for i, result in enumerate(pool.imap_unordered(process_dp, design_points)):
            proc_result(outs_sch, times, solve_times, result)  
            sys.stderr.write('\rdone {0:%}'.format(i/len(design_points)))
        pool.close()
    except Exception as e:
        print "hit an exception"
        print e
        pool.terminate()
        pool.join()

def titlefor(headers):
    return str(headers['cores']) + '--' + str(headers['way_gran'])

def main():
    parser = argparse.ArgumentParser()
    # parser.add_argument('-d', "--datafile", type = argparse.FileType('w'),
                        # default = sys.stdout,
                        # help = "File to output csv file to")
    parser.add_argument('-p', "--plot", action='store_true',
                        help = "For plotting generated PETs")
    parser.add_argument('-m', "--processors", default=cpu_count(), type = int,
                        help = "Number of processors to execute on")
    args = parser.parse_args()
                        
    exp = storage()

    exp.task_util = NAMED_UTILIZATIONS.keys()
    exp.inflation_A = NAMED_A_INFLATIONS.keys()
    exp.period = NAMED_PERIODS.keys()
    exp.memory_util = NAMED_MEMORY.keys()
    exp.util_dist = NAMED_UTIL_DISTS.keys()
    exp.sys_util = myrange(1.0, 37.0, 0.4)
    exp.cores = [32.0]#[4.0, 8.0, 16.0, 32.0]
    exp.way_gran = [4.0]#[1.0, 2.0, 4.0]
    
    design_points = [dp for dp in DesignPointGenerator(exp, is_valid = valid)]

    random.seed(43215)
    # random.shuffle(design_points)
    
    loadohs.load("../oheads/")

    print "Total design points: ", len(design_points)

    outs_sch = dict()
    outs_tme = dict()
    outs_slv = dict()
    manager = Manager()
    times = manager.dict()
    solve_times = manager.dict()
    for dp in design_points:
        title = titlefor(dp)
        if not title in outs_sch:
            outs_sch[title] = DictWriter(open("../scale_sched/" + title, 'w'), ['sys_util', 'sched'])
            outs_tme[title] = DictWriter(open("../scale_time/" + title, 'w'), exp.sys_util)
            outs_slv[title] = DictWriter(open("../scale_solve/" + title, 'w'), exp.sys_util)
            outs_sch[title].writeheader()
            outs_tme[title].writeheader()
            outs_slv[title].writeheader()
            times[title] = manager.list([manager.dict() for i in range(MAX_SAMPLES)])
            solve_times[title] = manager.list([manager.dict() for i in range(MAX_SAMPLES)])
    
#     print "before", times    
    
#     for i, dp in enumerate(design_points):
#         result = process_dp(dp)
#         proc_result(outs_sch, times, solve_times, result)  
     
    thread_tests(args, outs_sch, times, solve_times, design_points)

#     print "after", times     
    for title in times.keys():
        for row in times[title]:
            outs_tme[title].writerow(row)
    for title in solve_times.keys():
        for row in solve_times[title]:
            outs_slv[title].writerow(row)
            


if __name__ == '__main__':
    main()